/*
 * gencall.s
 *				    Gianni Mariani 3-Dec-1993
 *
 *  Generalized calling function.  This allows the calling
 *  of any function adhering to the MIPS calling conventions.
 *  It manipulates the CPU registers to adhere to the calling
 *  standards.  This allows high level ruotines to call any
 *  function at run time by assembling the parameters in the
 *  stack.
 *
 * CallFromGeneralArgs(
 *     preparefunc     Prefunc,	    / * Address of call args prep func	* /
 *     void          * args,	    / * Parameter for Prefunc		* /
 *     int             bytes,	    / * Parameter for Prefunc		* /
 *     int             flags,	    / * passing and return value flags  * /
 *     void          * retval,	    / * pointer to return value - opt.  * /
 *     ifunc_t         func	    / * address of function to call     * /
 * );
 *
 * Prefunc(
 *     char	     * call_args,   / * Address of first arg to func	* /
 *     void	     * args,	    / * As passed in CallFromGeneralArgs* /
 *     int	       bytes,	    / * As passed in CallFromGeneralArgs* /
 *     int	     * p_flags,	    / * Pointer to the value "flags"	* /
 *				    / * passed in CallFromGeneralArgs	* /
 * );
 *
 */

#define GCIF_ASHEADER
#include "gencall.h"

#include <asm.h>
#include <regdef.h>   

#define Prefunc	    a0
#define args	    a1
#define bytes	    a2
#define flags	    a3

#define SZINT	    4
#define SZSHORT	    2

#define FRM_SIZ	    ( 4 * SZINT + 2 * SZINT )

	.globl	CallFromGeneralArgs
	.ent	CallFromGeneralArgs 2
CallFromGeneralArgs:

	subu	sp, FRM_SIZ
	sw	fp, FRM_SIZ - 2*SZINT(sp)
	sw	ra, FRM_SIZ - 1*SZINT(sp)
	move	fp, sp
	.mask	0xc0000000, - 1*SZINT
	.frame	fp, FRM_SIZ, ra
	
	move	t9, Prefunc
	sw	flags, FRM_SIZ + 3*SZINT(fp)

 #  Allocate at least 16 bytes in the argstack
 #  argstack = alloca( bytes < 16 ? 16 : bytes );

	move	v0, bytes
	bge	bytes, 4 * SZINT, 33f
	li	v0, 4 * SZINT
	b	34f			# no need to round v0
33:
	addu	t0, v0, 2 * SZINT -1	# make sure it is aligned to 2*SZINT
	and	v0, t0, -2 * SZINT	# boundaries
34:
	subu	sp, sp, v0		# move the stack pointer to reflect
					# the arg space
	
 #  if ( Prefunc ) {
 #  	( * Prefunc )( argstack, args, bytes, &flags );

	beqz	t9, 34f
					# set up regs to call Prefunc
					# argstack is up sp + 4 * SZINT
					# args is already in a2
					# bytes is already in a3
					# &flags is fp + FRM_SIZ + 3*SZINT
	addu	a0, sp, 4 * SZINT
	addu	a3, fp, FRM_SIZ + 3*SZINT
	
	jal	t9
	
34:

 #  This is where we do special assembler stuff to arrange the call stack
docall:

	lh	t0, FRM_SIZ + 3*SZINT(fp)  # upper hw of flags
					   # for PASSFPREG

	addu	sp, sp, 4 * SZINT	# adjust sp to new args

 # Load args
	
 #  if ( ! t0 = ( flags & 0xffff0000 ) ) {
	bnez	t0, 38f			# make it quick for int
	lw	a0, 0*SZINT(sp)		# just go ahead and load the
	lw	a1, 1*SZINT(sp)		# four regs.
	lw	a2, 2*SZINT(sp)
	lw	a3, 3*SZINT(sp)
	b	20f
38:
	bne	t0, PASS_D, 39f
	l.d	$f12, 0*SZINT(sp)	# load fp regs from args
	lw	a2,   2*SZINT(sp)	# passing a double
	lw	a3,   3*SZINT(sp)
	b	20f
39:
	bne	t0, PASS_F, 40f
	l.s	$f12, 0*SZINT(sp)	# load fp regs from args
	lw	a1,   1*SZINT(sp)	# passing a float
	lw	a2,   2*SZINT(sp)
	lw	a3,   3*SZINT(sp)
	b	20f		
40:
	bne	t0, PASS_D_D, 41f
	l.d	$f12, 0*SZINT(sp)	# load fp regs from args
	l.d	$f14, 2*SZINT(sp)	# passing two doubles
	b	20f
41:
	bne	t0, PASS_F_F, 42f
	l.s	$f12, 0*SZINT(sp)	# load fp regs from args
	l.s	$f14, 1*SZINT(sp)	# passing two floats
	lw	a2,   2*SZINT(sp)
	lw	a3,   3*SZINT(sp)
	b	20f
42:
	bne	t0, PASS_D_F, 43f
	l.d	$f12, 0*SZINT(sp)	# load fp regs from args
	l.s	$f14, 2*SZINT(sp)	# passing double and float
	lw	a3,   3*SZINT(sp)
	b	20f
43:
 # assume that the only other combination must be float then double
 #	bne	t0, PASS_F_D, 20f
	l.s	$f12, 0*SZINT(sp)	# load fp regs from args
	l.d	$f14, 2*SZINT(sp)	# passing double and float

20:

 #  Make sure that the retval is set
 #  if ( ! retval ) goto noretval;

	lw	t9, FRM_SIZ + 5*SZINT(fp)
	lw	t1, FRM_SIZ + 4*SZINT(fp)
	beqz	t1, noretval
	
 #  Get the return value type
 #  flags = * ( 0 + ( short * ) & flags );

	lh	t2, FRM_SIZ + 3*SZINT + SZSHORT(fp)

 #  if ( flags == RETINT ) {

	bne     t2, RETINT, 25f
	jal	t9
	lw	t0, FRM_SIZ + 4*SZINT(fp)
	sw	v0, 0(t0)
	b	28f

 #  } else if ( flags == RETFLOAT ) {
25:
	bne     t2, RETFLOAT, 26f
	jal	t9
	lw	t0, FRM_SIZ + 4*SZINT(fp)
	s.s	$f0, 0(t0)
	b	28f

 #  } else if ( flags == RETDOUBLE ) {
26:
	bne	t2, RETDOUBLE, 27f
	jal	t9
	lw	t0, FRM_SIZ + 4*SZINT(fp)
	s.d	$f0, 0(t0)
	b	28f
27:
noretval:

	jal	t9
28:

	move	sp, fp
	lw	fp, FRM_SIZ - 2*SZINT(sp)
	lw	ra, FRM_SIZ - 1*SZINT(sp)
	addu	sp, FRM_SIZ
	j	ra

	END( CallFromGeneralArgs )

