/*
 * Copyright 1991, 1992, 1993, 1994, Silicon Graphics, Inc.
 * All Rights Reserved.
 *
 * This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
 * the contents of this file may not be disclosed to third parties, copied or
 * duplicated in any form, in whole or in part, without the prior written
 * permission of Silicon Graphics, Inc.
 *
 * RESTRICTED RIGHTS LEGEND:
 * Use, duplication or disclosure by the Government is subject to restrictions
 * as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
 * and Computer Software clause at DFARS 252.227-7013, and/or in similar or
 * successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
 * rights reserved under the Copyright Laws of the United States.
 */

#include <gl.h>
#include <device.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/fcntl.h>
#include <sys/errno.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <netdb.h>
#include <time.h>
#include <bstring.h>
#ifdef __cplusplus
#include <osfcn.h>
#endif
#include "gizmo.h"

long 	gizllx, gizlly;
long  	gid = 0, subid, subid2 = 0, sock, stowed;
long	colmode;
short	inited=0;
short	indigo = 0;
static	long	lastmx; 
static	long	lastmy;
extern	gfblock_t	*gfb;
extern	gizmonumber;

void createsocket(long);
void sendpair(long, long);
void switchgizmowindow(long);
void gizmo();
void drawsubframe();
void drawgizmo();
void readfromsocket();
void closewindow();
long readlongfromsocket();
void sendsocket(char *, long);

long global_argc;
char **global_argv;

#define UIBUTTONDOWN	1
#define UIBUTTONCLICK	2

void clearbackground()
{
    color(WHITE); clear();
}

void gizmoswapbuffers()
{
    swapbuffers();
    /* uiswapbuffers(); */
}

main(int argc, char **argv)
{
    int		portnum;

    /*gettmpdir();*/
    portnum = atoi(argv[1]);
    colmode = atoi(argv[2]);
    gizllx = atoi(argv[3]);
    gizlly = atoi(argv[4]);
    global_argc = argc;
    global_argv = argv;
    
    if (portnum) {
	createsocket(portnum);
    } else sock = -1;
    gizmo();
}

long
initwindow(long width, long height)
{
    long gid;

    foreground();
    prefposition(gizllx, gizllx+width-1, gizlly, gizlly+height-1);
    gid = winopen(global_argv[0]);
    wintitle(gfb->name);
    icontitle(gfb->name);

    if (!colmode) {
	if (getgdesc(GD_BITS_NORM_DBL_RED) == 1)
	    indigo = 1;
	else
	    RGBmode();
    }

    doublebuffer();

    glcompat(GLC_MQUEUERATE, GLC_COMPATRATE);
    gconfig();
    clearbackground();
    gizmoswapbuffers();
    getorigin(&gizllx, &gizlly);
    stowed = 0;
    icontitle(gfb->iconname);
    return gid;
}

void gizmo()
{
    short val;
    static long mx, my, dev;
    short smx, smy;
    fd_set	fdset;
    long	gd;
    short	redrawcount = 0;

    gid = initwindow((*gfb->getwidth)(), (*gfb->getheight)());
    (*gfb->initwidgets)();
    if (gfb->initsubwindow) {
    	subid = (*gfb->initsubwindow)();
    	switchgizmowindow(gid);
    } else
	subid = 0;
    if (subid) redrawcount = 2;
    if (subid2) redrawcount = 5;
    qdevice(LEFTMOUSE);
    qdevice(KEYBD);
    qdevice(MOUSEX);
    qdevice(MOUSEY);
    qdevice(WINQUIT);
    qdevice(WINSHUT);
    qdevice(LEFTARROWKEY);
    qdevice(RIGHTARROWKEY);
    qdevice(INPUTCHANGE);
    qdevice(WINFREEZE);
    qdevice(WINTHAW);
    tie(LEFTMOUSE, MOUSEX, MOUSEY);
    (*gfb->activatewidgets)();
    drawsubframe();
    if (sock == -1) {	/* debugging */
	/*initui();*/
	drawgizmo();
    }

#ifndef V3_3
    gd = qgetfd();
    if(gd <0) {
	perror("qgetfd failed");
	exit(1);
    }
#else
    if ((gd = open("/dev/queue",O_RDONLY)) < 0) {
	perror("opening /dev/queue");
	exit(1);
    }

#endif
    while (1) {
	FD_ZERO(&fdset);
	FD_SET(gd,&fdset);
	if (sock >= 0 )
	    FD_SET(sock,&fdset);
	select(FD_SETSIZE, &fdset, 0, 0, 0);
	if ( (sock >= 0) && (FD_ISSET(sock,&fdset))) {
	    readfromsocket();
	}
	if (!FD_ISSET(gd,&fdset))
	    continue;
	if (!gid){ /*window is stowed*/
	    qreset();
	    continue;
	}
	while (qtest())	{
	    if (!gid) {
		qread(&val);
		continue;
	    }
	    dev = qread(&val);
	    switch (dev) {
		case KEYBD:
		    if ((*gfb->interpchar)((long) val))
			if (inited) drawgizmo();
		    break;
		case MIDDLEMOUSE:
		case LEFTMOUSE:
		    qread(&smx); qread(&smy);
		    mx = smx - gizllx;
		    my = smy - gizlly;
		    if (val == UIBUTTONDOWN) {
			sginap(8);
			if (qtest() == dev) {
			    short tmp;
			    qread(&tmp); qread(&tmp); qread(&tmp);
			    val = UIBUTTONCLICK;
			} 
		    } 
		    if ((*gfb->interpclick)(mx, my, dev, (long) val))
			if (inited) 
			    drawgizmo(); 
		    break;
		case MOUSEX:
		    mx = val - gizllx;
		    break;
		case MOUSEY:
		    my = val - gizlly;
		    if ((mx != lastmx) || (my != lastmy)) {
			(*gfb->locatemouse)(mx, my);
			lastmx = mx;
			lastmy = my;
		    }
		    break;
		case REDRAW:
		    getorigin(&gizllx, &gizlly);
		    if (redrawcount) {
			redrawcount--;
			break;
		    }
		    sendpair(GIZMOXORG, (gizllx << 8 ) | gizmonumber);
		    sendpair(GIZMOYORG, (gizlly << 8 ) | gizmonumber);
		    reshapeviewport();
		    (*gfb->interpother)(dev, (long) val);
		    if (inited) drawgizmo();
		    if (inited) drawsubframe();
		    break;
		case WINQUIT:
		case WINSHUT:
		    sendpair(QUITGIZMO, gizmonumber);
		    /* return 2 from interpother if you don't want gizmo 
		       to exit */
		    if ( ((*gfb->interpother)(dev, (long) val))  != 2) {
			gexit();
			exit(1);
		    }
		    break;
		case WINTHAW:
		    stowed = 0;
		    if (inited) drawgizmo();
		    if (inited) drawsubframe();
		    break;
		case WINFREEZE:
		    stowed = 1;
		    if ((*gfb->interpother)(dev, (long) val)) {
			if (inited) drawgizmo();
			if (inited) drawsubframe();
		    }
		    break;
		case INPUTCHANGE:
		    /*if (!val)
			unlocateall();*/
		default:
		    if ((*gfb->interpother)(dev, (long) val))
			if (inited) drawgizmo();
		    break;
	    }
    	}
    }
}

void readfromsocket()
{
    long	cmd;

    if (sock == -1) return;
    cmd = readlongfromsocket();
    switch (cmd) {
	/* Standard */
	case WINTHAW:
    	    gid = initwindow((*gfb->getwidth)(), (*gfb->getheight)());
    	    if (gfb->initsubwindow) {
    		subid = (*gfb->initsubwindow)();
    		switchgizmowindow(gid);
    	    } else
		subid = 0;
	    /*initui();*/
	    inited = 1;
	    drawgizmo();
	    drawsubframe();
	    (*gfb->interpother)(cmd, 0);
	    break;
	case WINFREEZE:
	    if (subid) winclose(subid);
	    if (subid2) winclose(subid2);
	    closewindow();
	    return;
	case POKEGIZMO:
	    if (stowed) {
	    	if (subid) winclose(subid);
	    	if (subid2) winclose(subid2);
	    	winclose(gid);
    	    	gid = initwindow((*gfb->getwidth)(), (*gfb->getheight)());
    	    	if (gfb->initsubwindow) {
    		    subid = (*gfb->initsubwindow)();
    		    switchgizmowindow(gid);
    	        } else
		    subid = 0;
		drawgizmo();
	    	drawsubframe();
	    	(*gfb->interpother)(WINTHAW, 0);
	    } else
	    if (inited) {
	        winpop();
    	        drawgizmo();
	    	drawsubframe();
	    }
	    break;

	/* all the initializations have been sent to the gizmo */
	case GIZMOINITEND:
	    inited = 1;
    	    /*initui();*/
    	    drawgizmo();
	    drawsubframe();
	    break;
	default:
	    if ((*gfb->readfromsocket)(cmd))
		if (inited)
    		    drawgizmo();
	    break;
    }
}

void closewindow()
{
    if (gid) winclose(gid);
    gid = 0;
    inited = 0;
}

void drawgizmox()
{
    clearbackground();
    /* draw stuff */
    (*gfb->drawwidgets)();
    gizmoswapbuffers();
}

void drawgizmo()
{
    if (!gid || stowed) return;
    drawgizmox();
    if (colmode != 8) {
    	readsource(SRC_FRONT);
    	rectcopy(0, 0, (short)((*gfb->getwidth)()-1),
	    	    (short)((*gfb->getheight)()-1), 0, 0);
    	readsource(SRC_AUTO);
    }
}

void drawsubframe()
{
    long id;

    if (!subid) return;
    id = winget();

    switchgizmowindow(subid);

    /* draw stuff */
    (*gfb->drawsubwindow)();

    if (id == gid)
    	switchgizmowindow(gid);
}

void sendmemmessage(char *message,long msgtype)
{
long strlength;

	sendpair(MESSAGEDEV,gizmonumber);
	sendsocket((char *)&msgtype, sizeof(long));
	strlength = strlen(message);
	sendsocket((char *)&strlength, sizeof(long));
	sendsocket(message, strlength);
}

void assertfunc(long x, char *str)
{
    if (x) return;
    perror(str);
    exit(1);
}

void switchgizmowindow(long towindow)
{
    long id = winget();
    long xsize, ysize;

    if (towindow == id) return;

    winset(towindow);

    getsize(&xsize, &ysize);

    reshapeviewport();
    ortho2(-0.5, xsize-0.5,-0.5, ysize-0.5);
}
